/*
 *  Copyright (C) 2002-2003  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "mt32emu.h"
#include "mixer.h"

#include <SDL_Thread.h>

MT32Emu::Synth *_usesynth;
MixerChannel *mt32chan;

static struct {
    SDL_mutex * mutex;
    SDL_Thread * thread;
    SDL_semaphore * sem;
    volatile bool running, busy;

    Bit8u len, play;
    Bit8u Temp[MIXER_BUFSIZE], msg[SYSEX_SIZE];
} mt32;

static int MT32_Thread(void*) {

   SDL_LockMutex(mt32.mutex);
   while(mt32.running) {

      Bitu len;
      if(!(mt32.play) && !(mt32.len)) {
        SDL_UnlockMutex(mt32.mutex);
        SDL_SemWait(mt32.sem);
        SDL_LockMutex(mt32.mutex);
      }

      mt32.busy = true;
      while(mt32.play) {

	    len = mt32.play;
	    Bit32u *tmp = ((Bit32u*)mt32.msg);
	    SDL_UnlockMutex(mt32.mutex);
	    while(len--) {
		_usesynth->playMsg(*tmp++);
	    }
	    SDL_LockMutex(mt32.mutex);
	    len = tmp-(Bit32u*)mt32.msg;
	    mt32.play -= len;
	    if(mt32.play) SDL_memmove(mt32.msg, tmp, mt32.play<<2);
      }

      while(mt32.len) {

	    len = (mt32.len>>2) < MIXER_BUFSIZE ? mt32.len : MIXER_BUFSIZE<<2;
	    mt32.len -= len;

#ifdef MT32DEBUG
	    if(mt32.len) LOG_MSG("MT32:WARNING: len left (%d)", mt32.len);
#endif

	    SDL_UnlockMutex(mt32.mutex);
	    _usesynth->render((Bit16s *)mt32.Temp, len);
	    mt32chan->AddSamples_s16(len,(Bit16s *)mt32.Temp);
	    SDL_LockMutex(mt32.mutex);
	    break;
      }
      mt32.busy = false;
   }

   SDL_UnlockMutex(mt32.mutex);
   return 0;
}

static void MT32_CallBack(Bitu len) {
   SDL_LockMutex(mt32.mutex);
   mt32.len += len;
   SDL_UnlockMutex(mt32.mutex);
   SDL_SemPost(mt32.sem);
}

static int report(void *userData, MT32Emu::ReportType type, const void *reportData) {
   switch(type) {
   case MT32Emu::ReportType_errorControlROM:
      LOG(LOG_ALL,LOG_ERROR)("MT32:Couldn't find control files");
      break;
   case MT32Emu::ReportType_errorPCMROM:
      LOG(LOG_ALL,LOG_ERROR)("MT32:Couldn't open MT32_PCM.ROM file");
      break;
   default:
      //LOG(LOG_ALL,LOG_NORMAL)("MT32: Report %d",type);
      break;
   }
   return 0;
}

class MidiHandler_mt32: public MidiHandler {
private:
   MT32Emu::Synth *_synth;
   int _outputRate;
   bool isOpen;

public:
   MidiHandler_mt32() : isOpen(false),MidiHandler() {};
   const char * GetName(void) { return "mt32";};
   bool Open(const char * conf) {
      MT32Emu::SynthProperties tmpProp;
      memset(&tmpProp, 0, sizeof(tmpProp));
      tmpProp.sampleRate = atoi(conf);
      if(tmpProp.sampleRate < 8192) tmpProp.sampleRate = 22050;
      LOG_MSG("MT32:Set sample rate to %d", tmpProp.sampleRate);

      /* Create MT32 thread */
      mt32.mutex = SDL_CreateMutex();
      mt32.sem = SDL_CreateSemaphore(0);
      mt32.running = true;
      mt32.busy = false;
      mt32.play = 0;
      mt32.thread = SDL_CreateThread(MT32_Thread, NULL);

      tmpProp.useDefaultReverb = true;
      tmpProp.useReverb = true;
      tmpProp.reverbType = 1;
      tmpProp.reverbTime = 5;
      tmpProp.reverbLevel = 3;
      //tmpProp.printDebug = &vdebug;
      tmpProp.report = &report;
      _synth = new MT32Emu::Synth();
      if (_synth->open(tmpProp)==0) {
         LOG(LOG_ALL,LOG_ERROR)("MT32:Error initialising emulation");
         return false;
      }
      _usesynth=_synth;

      mt32chan=MIXER_AddChannel(MT32_CallBack,tmpProp.sampleRate,"MT32");
      mt32chan->Enable(false);
      return true;
   };
   void Close(void) {
      if (!isOpen) return;
      mt32chan->Enable(false);
      mt32.running = false;
      SDL_SemPost(mt32.sem);
      SDL_WaitThread(mt32.thread, NULL);
      SDL_DestroyMutex(mt32.mutex);
      _synth->close();
      delete _synth;
      _synth = NULL;
      isOpen=false;
   };
   void PlayMsg(Bit8u * msg) {
      mt32chan->Enable(true);

      // Try to queue play commands
      SDL_LockMutex(mt32.mutex);
      // Playcommand buffer full?
      while(!(mt32.play < SYSEX_SIZE>>2)) {
        SDL_UnlockMutex(mt32.mutex);
        LOG_MSG("MT32:Playback buffer full...");
        SDL_LockMutex(mt32.mutex);
      }
      SDL_memcpy(mt32.msg+(mt32.play<<2), msg, sizeof(Bit32u));
      mt32.play ++;
      SDL_UnlockMutex(mt32.mutex);
      SDL_SemPost(mt32.sem);
   };
   void PlaySysex(Bit8u * sysex,Bitu len) {
      SDL_LockMutex(mt32.mutex);
      while(mt32.busy) {
        SDL_UnlockMutex(mt32.mutex);
#ifdef MT32DEBUG
        LOG_MSG("MT32:Waiting to deliver sysex");
#endif
        SDL_LockMutex(mt32.mutex);
      }

      if(sysex[0] == 0xf0) {
         _synth->playSysex(sysex, len);
      } else {
         _synth->playSysexWithoutFraming(sysex, len);
      }
      SDL_UnlockMutex(mt32.mutex);
   };
};

MidiHandler_mt32 Midi_mt32;
